""" Note:

Serve vLLM using the command

docker run --gpus all -p 8000:8000 vllm/vllm-openai:latest --model Qwen/Qwen3-1.7B --trust-remote-code --gpu-memory-utilization 0.8 --max_model_len 8192 --enable-auto-tool-choice --tool-call-parser hermes

Since vLLM preallocates GPU and can't offload to CPU, this is a safe command for testing using a 8Gb RAM GPU.
"""
from openai import OpenAI
from .openai import LLM_GPT_OpenAI


class LLM_VLLM(LLM_GPT_OpenAI):
    def __init__(self, model):
        """Constructor
        Arguments:
            model - Desired model
        """
        # there would be electricity cost for running a model locally
        # but for now we keep cost at 0
        self.price_per_M_input_tokens = 0
        self.price_per_M_output_tokens = 0
        if model == "Qwen 3 1.7b VLLM":
            self.model_id = "Qwen/Qwen3-1.7B"
            self.llm_description = "Qwen 3 1.7b (Tiny-size LLM) - locally from vLLM"

        try:
            self.openai_client = OpenAI(
                base_url="http://localhost:8000/v1/",
                api_key="vllm",  # required but ignored
            )
        except Exception:
            self.openai_client = None

        self.config = {
            # "messages": prompt,
            # "system": sysprompt,
            "max_tokens": 2048,
            "temperature": 0.5,  # 0.5 is default,
            "stream": True,
            # "top_k": 250,
            # "top_p": 1,
            "stop": None,  # the regular is already implemented
            "model": self.model_id,
        }
        # requests and answer word count
        self.word_counts = []



""" Handles the connection with MCP servers

More details about the config parameters:
https://gofastmcp.com/clients/transports#mcp-json-configuration-transport

## TODO:
[ ] Call resources
[ ] Call prompts

The expected format is as follows:

config = {
    "mcpServers": {
        "server_name": {
            # Remote HTTP/SSE server
            "transport": "http",  # or "sse"
            "url": "https://api.example.com/mcp",
            "headers": {"Authorization": "Bearer token"},
            "auth": "oauth"  # or bearer token string
        },
        "local_server": {
            # Local stdio server
            "transport": "stdio",
            "command": "python",
            "args": ["./server.py", "--verbose"],
            "env": {"DEBUG": "true"},
            "cwd": "/path/to/server",
        }
    }
}
"""


import asyncio
from fastmcp import Client
from fastmcp.prompts.prompt import TextContent


class MCPTool:
    def __init__(self, name, description, call_fn):
        self.name = name
        self.tool_description = description
        self.tool_summary = description
        self.call_fn = call_fn

    def __call__(self, **args):
        return asyncio.run(self.call_fn(args))


class MCPConnector:
    @classmethod
    async def create_from_cfg(cls, config):
        client = Client(config)
        async with client:
            # Basic server interaction
            await client.ping()

            # List available operations
            try:
                tools = await client.list_tools()
            except:
                tools = None

            try:
                resources = await client.list_resources()
            except:
                resources = None

            try:
                prompts = await client.list_prompts()
            except:
                prompts = None

        mcpc = cls(config, client, tools, resources, prompts)
        return mcpc

    def __init__(
        self,
        config,
        client,
        mcp_tools,
        resources,
        prompts,
        return_only_text_content=True,
    ):
        """Constructor.
        Test connections, list available tools and instantiates their call process

        If return_only_text_content = True, only returns the textual part of the tool calls
        """
        self.config = config
        self.client = client
        self.mcp_tools = mcp_tools
        self.resources = resources
        self.prompts = prompts
        self.return_only_text_content = return_only_text_content
        self.tools = []
        for mcp_tool in self.mcp_tools:
            cur_tool_desc = self._convert_mcp_tool_to_std(mcp_tool)

            self.tools.append(
                MCPTool(
                    cur_tool_desc["name"],
                    cur_tool_desc,
                    self.call_tool(cur_tool_desc["name"]),
                )
            )

    def call_tool(self, name):
        async def call_tool_fn(args):
            ans = await self._call_tool_async(name, args)
            return ans

        return call_tool_fn

    async def _call_tool_async(self, name, args):
        async with self.client:
            # Simple tool call
            result = await self.client.call_tool(name, args)
        if self.return_only_text_content:
            ans = [str(x.text) for x in result.content if isinstance(x, TextContent)]
            if len(ans) == 1:
                return str(ans[0])
            elif len(ans) == 0:
                return ""
            else:
                return str(ans)
        else:
            return str(result.content)

    def _convert_mcp_tool_to_std(self, mcp_tool_description):
        """Converts a MCP tool description to Claude's standard format"""
        tool_desc = {
            "name": mcp_tool_description.name,
            "description": mcp_tool_description.description,
            "input_schema": mcp_tool_description.inputSchema,
        }
        if tool_desc["description"] is None:
            # handle more gracefully
            tool_desc["description"] = "(no description provided)"
        tool_desc["input_schema"]["type"] = "object"
        tool_desc["input_schema"]["required"] = tool_desc["input_schema"].get(
            "required", []
        )
        return tool_desc


def main():
    config = {
        "mcpServers": {
            "calc2": {
                # Remote HTTP/SSE server
                "transport": "http",  # or "sse"
                "url": "http://127.0.0.1:8000/mcp",
                # "headers": {"Authorization": "Bearer token"},
                # "auth": "oauth"  # or bearer token string
            },
            # "huggingface": {
            #    "transport": "http",
            #    "url": "https://huggingface.co/mcp",
            # }
        }
    }
    config = {
        "mcpServers": {
            "huggingface": {
                "url": "https://huggingface.co/mcp",
                # "headers": {"Authorization": "Bearer TOKEN"}
            }
        }
    }
    mcpc = asyncio.run(MCPConnector.create_from_cfg(config))
    for t in mcpc.tools:
        print(t.name, t.tool_description)
        if t.name == "hf_doc_search":
            break
    # print(mcpc.tools[3].name, mcpc.tools[3].tool_description)
    # args = {"a": 5, "b": 4}
    args = {"query": "text to speech"}
    tool_ans = t(**args)
    print(t.name, tool_ans)


if __name__ == "__main__":
    # python -m gat_llm.connector_mcp
    main()



import os

import numpy as np
import pandas as pd
from openai import OpenAI

try:
    import parselmouth
except ImportError:
    parselmouth = None


rng = np.random.default_rng()


def fill_zeros_with_last(arr):
    prev = np.arange(len(arr))
    prev[arr <= 0] = 0
    prev = np.maximum.accumulate(prev)
    return arr[prev]


def combine_transcription_with_speech_params(transcription_dict, speech_params):
    """
    Combines the transcription with speech parameters

    Args:
        transcription_dict: OpenAI style transcript dict with key "words" and a list of type
            'words': [{'end': 2.4, 'start': 2.3, 'word': 'Hello'},
        speech_params: dictionary with keys time_s and the various voice parameters (see analyze_voice)
    """
    # accumulate the values to make it efficient to compute averages
    accum_params = {}
    delta_t = speech_params["time_s"][1] - speech_params["time_s"][0]
    for k in speech_params:
        accum_params[f"accum_{k}"] = (
            np.cumsum(fill_zeros_with_last(np.nan_to_num(speech_params[k]))) * delta_t
        )
    accum_params["time_s"] = speech_params["time_s"]

    # extract the words and their start / stop times
    word_start_stops = np.array(
        [[x["start"], x["end"]] for x in transcription_dict["words"]]
    )
    word_durations = word_start_stops[:, 1] - word_start_stops[:, 0]

    all_words = [
        {
            "word": x["word"],
            "start_time_s": float(np.round(x["start"], decimals=1)),
            "end_time_s": float(np.round(x["end"], decimals=1)),
        }
        for x in transcription_dict["words"]
    ]

    for k in speech_params:
        if k != "time_s":
            cur_avg = np.interp(
                word_start_stops, accum_params["time_s"], accum_params[f"accum_{k}"]
            )
            cur_avg = np.divide(cur_avg[:, 1] - cur_avg[:, 0], word_durations + 1e-5)
            for i in range(len(all_words)):
                all_words[i][k] = float(np.round(cur_avg[i], decimals=1))

    df = pd.DataFrame(all_words)
    return df.to_xml(
        index=False,
        xml_declaration=False,
        root_name="speech_and_features",
        row_name="word_and_features",
    )


def analyze_voice(
    input_wav, frame_step=0.01, window=0.2, pitch_floor=75, pitch_ceiling=450
):
    """Analyzes voice using Praat.
    Note that it would be better to customize pitch floor and ceiling to gender / children
    """
    snd = parselmouth.Sound(input_wav)

    # === Fundamental frequency (F0) ===
    pitch = snd.to_pitch(
        time_step=frame_step, pitch_floor=pitch_floor, pitch_ceiling=pitch_ceiling
    )
    f0_values = pitch.selected_array["frequency"]  # Hz, 0 = unvoiced
    f0_times = pitch.xs()

    # === Harmonicity (HNR) ===
    harmonicity = snd.to_harmonicity_cc(time_step=frame_step, minimum_pitch=pitch_floor)
    hnr_values = harmonicity.values.flatten()  # 1D array
    hnr_times = harmonicity.xs()

    # === Intensity ===
    intensity = snd.to_intensity(time_step=frame_step, minimum_pitch=pitch_floor)
    intensity_values = intensity.values.flatten()  # dB
    intensity_times = intensity.xs()

    # === Point process (for jitter & shimmer) ===
    point_process = parselmouth.praat.call(
        snd,
        "To PointProcess (periodic, peaks)",
        pitch_floor,
        pitch_ceiling,
        "yes",
        "no",
    )

    jitters = []
    shimmers = []

    for t in f0_times:
        start = max(0, t - window / 2)
        end = min(snd.xmax, t + window / 2)
        try:
            jitter_local = parselmouth.praat.call(
                point_process, "Get jitter (local)", start, end, 0.0001, 0.02, 1.3
            )
            shimmer_local = parselmouth.praat.call(
                [snd, point_process],
                "Get shimmer (local)",
                start,
                end,
                0.0001,
                0.02,
                1.3,
                1.6,
            )
        except Exception:
            jitter_local = np.nan
            shimmer_local = np.nan

        jitters.append(jitter_local)
        shimmers.append(shimmer_local)

    # === Align HNR and Intensity to F0 time grid ===
    hnr_interp = np.interp(f0_times, hnr_times, hnr_values, left=np.nan, right=np.nan)
    intensity_interp = np.interp(
        f0_times, intensity_times, intensity_values, left=np.nan, right=np.nan
    )

    ans = {
        "time_s": f0_times,
        "f0_hz": f0_values,
        "hnr_db": hnr_interp,
        "intensity_db": intensity_interp,
        "jitter_percentage": [x * 100 for x in jitters],
        "shimmer_db": shimmers,
    }
    return ans


class ToolSpeechAnalysis:
    def __init__(self):
        self.name = "speech_analysis"

        self.tool_description = {
            "name": self.name,
            "description": """Uses an automatic speech recognition tool to convert the provided audio into text.
Uses Praat to analyze frequency, intensity, SNR and other speech parameters. Each word is provided along with its start time, duration, and
Returns: path to a file containing a JSON description of the audio, along with various useful parameters.
Raises ValueError: if not able to read the audio or video file.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "audio_file_path": {
                        "type": "string",
                        "description": "Path to the audio file that will be converted to text.",
                    },
                    "language": {
                        "type": "string",
                        "description": "Language of the audio in file audio_file_path. Should be provided in ISO-639-1 (e.g. en, pt, es) format. Defaults to en",
                    },
                },
                "required": ["audio_file_path"],
            },
        }
        self.openai_client = None

    def __call__(
        self,
        audio_file_path,
        language="en",
        return_path_to_file_only=True,
        **kwargs,
    ):
        """Creates a transcription of an audio file along with speech parameters.

        Arguments:
          - audio_file_path: Path to the file that should be transcribed
          - language: language in ISO format (e.g. en, pt, es, it)
          - return_path_to_file_only: if True, writes a transcription file and returns it. If false, returns the transcription itself
        Returns:
          - The audio srt file or the full transcription
        """
        if parselmouth is None:
            return "Praat is not installed. Please install with `pip install praat-parselmouth`"
        os.makedirs("media", exist_ok=True)
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        # create client
        if self.openai_client is None:
            self.openai_client = OpenAI()

        if not os.path.isfile(audio_file_path):
            return f"Audio file not found: {audio_file_path}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/transcript_{rng_num}.xml"
        try:
            with open(audio_file_path, "rb") as audio_file:
                transcript = self.openai_client.audio.transcriptions.create(
                    model="whisper-1",
                    file=audio_file,
                    response_format="verbose_json",
                    timestamp_granularities=["word"],
                )
            transcript_json = transcript.dict()

            # Use Praat to extract relevant information
            speech_params = analyze_voice(audio_file_path)
            xml_content = combine_transcription_with_speech_params(
                transcript_json, speech_params
            )

            if return_path_to_file_only:
                with open(target_file, "w", encoding="UTF-8") as f:
                    f.write(xml_content)
            else:
                return transcript
        except Exception as e:
            return f"Transcription was NOT generated.\nError description: {str(e)}"

        if not os.path.isfile(target_file):
            return "Error: Transcription file was not saved correctly."

        ans = ["<transcript>"]
        ans.append(f"<path_to_file>{target_file}</path_to_file>")
        ans.append("</transcript>")
        return "\n".join(ans)
